/*-
 * Copyright (c) 2005 Fredrik Lindberg. <fli at shapeshifter dot se>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#if __FreeBSD_version >= 500000
# include <stdint.h>
#else
# include <inttypes.h>
#endif
#include <string.h>
#include <sys/file.h>
#include <errno.h>

#include <png.h>

#include "mylibpng.h"

/*
 * Creates a pseudo-image and allocates memory
 * This function will return a 'pngimage'-type if
 * sucessfull and NULL upon failure
 * Failure in this function can only occur if memory can't be allocated.
 */
pngimage 
mypng_new(uint32_t width, uint32_t height, uint8_t depth) 
{
	pngimage pngimg = NULL;
	uint8_t real_depth;

	/* Real depth is number of bytes per pixel */
	real_depth = (depth / 8) * 3; /* We assume 8 bits per byte */

	if ((pngimg = malloc(sizeof(struct mylibpng_image_t))) != NULL) {
		pngimg->width = width;
		pngimg->height = height;
		pngimg->depth = depth;
		pngimg->real_depth = real_depth;

		/* Allocate data block to hold raw image data */
		if ((pngimg->data = malloc(width * height * real_depth)) == NULL) {
			/* We failed to allocate memory to raw image data,
			 * free previous allocated block and set pointer to NULL */
			free(pngimg);
			pngimg = NULL;
		}
		else {
			/* Clean our new data block */
			memset(pngimg->data, 255, width * height * real_depth);
		}
	}
	else {
		pngimg = NULL; /* Just in case */	
	}

	return pngimg;
}

/*
 * Destroys a 'pngimage' and free()s all allocated memory
 */
void
mypng_destroy(pngimage pngimg)
{
	if (pngimg != NULL) {
		/* Double check that we really have a data block */
		if (pngimg->data != NULL) {
			free(pngimg->data);
		}

		free(pngimg);
	}
}

/*
 * Write the raw data in 'pngimg' to file pointed to by 'path'
 */
int
mypng_write(pngimage pngimg, char *path)
{
	FILE *fp;

	unsigned char *p, *q;
	uint64_t rowbytes;

	png_infop info_ptr;
	png_structp png_ptr;

	if ((fp = fopen(path, "wb")) != NULL) {
		/* Get an exclusive lock on the file */
		flock(fileno(fp), LOCK_EX);

		/* Create our libpng write struct */
		png_ptr = png_create_write_struct(
			PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);

		/* Create info structure */
		info_ptr = png_create_info_struct(png_ptr);

		/* Initiate I/O */
		png_init_io(png_ptr, fp);	

		/* Set up PNG header */
		png_set_IHDR(png_ptr, info_ptr, pngimg->width, pngimg->height,
			pngimg->depth, PNG_COLOR_TYPE_RGB, 
			PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT);	

		png_write_info(png_ptr, info_ptr);

		/* Start of data segment */
		p = pngimg->data; 

		/* End of data segment */
		q = p + pngimg->width * pngimg->height * pngimg->real_depth;

		/* Incase of stupid compilers */
		rowbytes = pngimg->width * pngimg->real_depth;

		/* Write RAW image data to png file */ 
		for (; p < q; p += rowbytes) {
			png_write_row(png_ptr, p);
		} 

		/* Finish image */
		png_write_end(png_ptr, NULL);

		/* Now clean up after us, image should have been written now */
		flock(fileno(fp), LOCK_UN);
		fclose(fp);
		png_destroy_write_struct(&png_ptr, &info_ptr);
	}
	else {
		return -1; 
	}

	return 0;
}

/*
 * Very primitive read function
 */
pngimage
mypng_read(char *path)
{
	pngimage pngimg = NULL;
	FILE *fp;
	char header[8];
	png_infop info_ptr;
	png_structp png_ptr;
	png_bytep * rowpointers;
	unsigned char *p;

	if ((fp = fopen(path, "rb")) == NULL) 
		return NULL; 

	fread(header, 1, 8, fp);
	if (png_sig_cmp(header, 0, 8)) {
		fclose(fp);
		errno = EFTYPE;
		return NULL;
	}

	/* Allocate private structures */
	if ((pngimg = malloc(sizeof(struct mylibpng_image_t))) == NULL) {
		fclose(fp);
		return NULL;
	}

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 
			(png_voidp)NULL, NULL, NULL);

	info_ptr = png_create_info_struct(png_ptr);

	/* Initiate I/O */
	png_init_io(png_ptr, fp);	

	/* Already read 8 bytes */
	png_set_sig_bytes(png_ptr, 8);

	/* Prepare the read */
	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);	

	pngimg->width = info_ptr->width;
	pngimg->height = info_ptr->height;
	pngimg->depth = info_ptr->bit_depth;
	pngimg->real_depth = (pngimg->depth / 8) * 3;

	if ((pngimg->data = malloc(pngimg->width * pngimg->height * pngimg->real_depth)) != NULL) {

		uint64_t rowbytes;
		uint32_t i;
		rowbytes = pngimg->width * pngimg->real_depth;

		rowpointers = png_get_rows(png_ptr, info_ptr);

		p = pngimg->data; 
		for (i = 0; i < pngimg->height; i++) {
			memcpy(p, rowpointers[i], rowbytes);	
			p += rowbytes;
		}
	}
	else {
		free(pngimg);
		pngimg = NULL;
	}

	fclose(fp);
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	return pngimg;
}

/*
 * Fill an entire image with a color
 */
void
mypng_fill8(pngimage pngimg, pngcolor8 *color)
{
	unsigned char *p = NULL, *q = NULL;
	if (color->r == color->g && color->g == color->b) {
		memset(pngimg->data, color->r, pngimg->width * pngimg->height * pngimg->real_depth);
	}
	else {
		q = p + (pngimg->width * pngimg->height * pngimg->real_depth); 
		for (p = pngimg->data; p < q; p++) {
			*p++ = color->r;
			*p++ = color->g;
			*p   = color->b;
		}
	}	
}

/*
 * Draw an arbitary large, filled box
 */
void
mypng_box8(pngimage pngimg, uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2, pngcolor8 *color) {

	unsigned char *p, *q, *r, *s;
	uint64_t rowbytes;

 	rowbytes = pngimg->width * pngimg->real_depth;

	p = pngimg->data + (y1 * rowbytes);
	q = pngimg->data + (y2 * rowbytes);

	if (x2 > x1 && y2 > y1) {

		for (; p < q; p += rowbytes) {
			r = p + (x2 * pngimg->real_depth);
			for (s = p + (x1 * pngimg->real_depth); s < r; s++) {
				*s++ = color->r;
				*s++ = color->g;
				*s   = color->b;
			}
		} 
	}
}

/*
 * Draw an arbitary, transparent box 
 * opacity=1 (100%) no transparency 
 * opacity=0.5 (50%) 50% transparency
 * opacity=0 (0%) not visible at all 
 */
void
mypng_box8_opacity(pngimage pngimg, uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2, pngcolor8 *color) {

	unsigned char *p, *q, *r, *s;
	uint64_t rowbytes;

 	rowbytes = pngimg->width * pngimg->real_depth;

	p = pngimg->data + (y1 * rowbytes);
	q = pngimg->data + (y2 * rowbytes);

	if (x2 > x1 && y2 > y1) {
		for (; p < q; p += rowbytes) {
			r = p + (x2 * pngimg->real_depth);
			for (s = p + (x1 * pngimg->real_depth); s < r; s++) {
				*s = (*s * (1 - color->opacity)) + (color->r * color->opacity);
				s++;
				*s = (*s * (1 - color->opacity)) + (color->g * color->opacity);
				s++;
				*s = (*s * (1 - color->opacity)) + (color->b * color->opacity);
			}
		} 
	}
}
/* 
 * Pseudo-plots a pixel onto a pseudo-image
 * THIS ROUTINE DOES NO ERROR CHECKING
 * to see if passed pixel coordinates are inside valid range.
 * This is a design choice to gain speed and not a flaw.
 *
 * If bongous coordinates are passed to this function 
 * you WILL end up writing data into unknown territory and the 
 * program will most likley crash with a segmentation fault. 
 *
 * However, if compiled with DEBUG an error message will be 
 * printed to stderr.
 */
void
mypng_plot8(pngimage pngimg, uint32_t x, uint32_t y, pngcolor8 *color)
{
	unsigned char *p;

#ifdef DEBUG
	/* If we were compiled with debug options, print some error
	 * messages if passed (x, y) is invalid */
	if ((x > (pngimg->width - 1)) || (y > (pngimg->height - 1))) {
		fprintf(stderr, "mypng_plot8(): Bongous coordinates (%u, %u) exceeds image size (%u, %u)\n", x, y, pngimg->width - 1, pngimg->height - 1); 
	}
#endif /* DEBUG */

	/* Set to start of data segment */
	p = pngimg->data;

	/* Add our 'pixel' offset */ 
	p += (y * pngimg->width * pngimg->real_depth) + (x * pngimg->real_depth); 

	/* Move pixel data into data segment */ 
	*p++ = color->r;
	*p++ = color->g;
	*p   = color->b;
}

/*
 * Pseudo-plots a pixel onto a pseudo-image
 * Same functionallity as mypng_plot8() except this one 
 * respects color->opacity
 */
void
mypng_plot8_opacity(pngimage pngimg, uint32_t x, uint32_t y, pngcolor8 *color)
{
	unsigned char *p;

#ifdef DEBUG
	/* If we were compiled with debug options, print some error
	 * messages if passed (x, y) is invalid */
	if ((x > (pngimg->width - 1)) || (y > (pngimg->height - 1))) {
		fprintf(stderr, "mypng_plot8(): Bongous coordinates (%u, %u) exceeds image size (%u, %u)\n", x, y, pngimg->width - 1, pngimg->height - 1); 
	}
#endif /* DEBUG */

	/* Set to start of data segment */
	p = pngimg->data;

	/* Add our 'pixel' offset */ 
	p += (y * pngimg->width * pngimg->real_depth) + (x * pngimg->real_depth); 

	/* Move pixel data into data segment */ 
	*p = (*p * (1 - color->opacity)) + (color->r * color->opacity);
	p++;
	*p = (*p * (1 - color->opacity)) + (color->g * color->opacity);
	p++;
	*p = (*p * (1 - color->opacity)) + (color->b * color->opacity);
}

/*
 * Find color for any given plotted pixel 
 * THIS ROUTINE DOES NO ERROR CHECKING
 * to see if passed pixel coordinates are inside valid range.
 * This is a design choice to gain speed and not a flaw.
 * 
 * Bongous coordinates will result in a segmentation fault
 */
void
mypng_get8(pngimage pngimg, uint32_t x, uint32_t y, pngcolor8 *color)
{
	unsigned char *p;

#ifdef DEBUG
	/* If we were compiled with debug options, print some error
	 * messages if passed (x, y) is invalid */
	if ((x > (pngimg->width - 1)) || (y > (pngimg->height - 1))) {
		fprintf(stderr, "mypng_get8(): Bongous coordinates (%u, %u) exceeds image size (%u, %u)\n", x, y, pngimg->width - 1, pngimg->height - 1); 
	}
#endif /* DEBUG */

	/* Set to start of data segment */
	p = pngimg->data;

	/* Add our 'pixel' offset */ 
	p += (y * pngimg->width * pngimg->real_depth) + (x * pngimg->real_depth); 

	/* Get pixel colors */
	color->r = *p++;
	color->g = *p++;
	color->b = *p;
}

